/*
 * Copyright (C) 2018-2021 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdbool.h>
#include <pthread.h>
#include <assert.h>
#include "mamba.h"

#include "mmb_tile.h"
#include "mmb_tile_iterator.h"
#include "mmb_logging.h"
#include "mmb_memory.h"
#include "i_mmb_hash.h"

/* Keep the full path to ease the research of the correpsonding header */
#include "../memory/allocation.h"
#include "../memory/strategy.h"
#include "../memory/memory_options.h"

#ifdef HAVE_LOOP_ANALYSIS
#include "mmb_loop.h"
#endif
#ifdef HAVE_COST_MODEL
#include "mmb_cost_model.h"
#else 
#define mmbCostModelHandle void*
#endif

/* protecting initialization */
static pthread_mutex_t g_mamba_state = PTHREAD_MUTEX_INITIALIZER;
static inline void mmb_state_lock(void) {
  int stat = pthread_mutex_lock(&g_mamba_state);
  assert(stat==0);
}
static inline void mmb_state_unlock(void) {
  int stat = pthread_mutex_unlock(&g_mamba_state);
  assert(0 == stat);
}
/* Types */
/* Internal singleton Mamba instance */
typedef struct s_mmbMambaInstance {
  bool is_initialized;
  mmbManager *manager;
  mmbCostModelHandle *cm_handle;
} mmbMambaInstance;
static mmbMambaInstance g_mmb_instance = {false, NULL, NULL};


static inline mmbError nd_to_1d_idx(mmbIndex *in_idx, mmbDimensions * in_dims, 
                                    size_t *out_idx){
  /* Validity and range checks */
  if(!in_idx || !in_dims || !out_idx)
    return MMB_INVALID_ARG;

  if(in_idx->size != in_dims->size){
    MMB_ERR("Index size %zu does not match dimensions size %zu\n", in_idx->size, in_dims->size);
    return MMB_ERROR; 
  }

  for(unsigned i = 0; i < in_idx->size; i++)
    if(in_idx->d[i] >= in_dims->d[i]) {
      MMB_ERR("Index %zu out of bounds for dimension %u size %zu\n", in_idx->d[i], i, in_dims->d[i]);
      return MMB_ERROR;
    }

  /* Convert to 1d */
  *out_idx = 0;
  for (unsigned i = 0; i < in_dims->size; i++) {
    size_t idx_i = in_idx->d[i];
    size_t sum_dims = 1;
    for(unsigned j = i+1; j < in_dims->size; j++) {
      sum_dims *= in_dims->d[i];
    }
    idx_i *= sum_dims;
    *out_idx += idx_i;
  }
  return MMB_OK;
}

mmbError mmb_init( mmbOptions *_in_opts ) {
  mmbError stat = MMB_OK;
  mmbOptions *in_opts = _in_opts;

  /* Check before we take the lock */
  if(g_mmb_instance.is_initialized) { 
    MMB_ERR("Mamba library already initialized\n");
    stat = MMB_ERROR;
    goto UNLOCK_BAILOUT;
  }

  mmb_state_lock();

  /* Check again in case someone beat us here */
  if(g_mmb_instance.is_initialized) { 
    MMB_ERR("Mamba library already initialized\n");
    stat = MMB_ERROR;
    goto UNLOCK_BAILOUT;
  }

  if(!_in_opts) {
    stat = mmb_options_create_default(&in_opts);
    if(stat != MMB_OK){
      MMB_ERR("Failed to create default mmb options\n");
      goto UNLOCK_BAILOUT;
    }
  }

  stat = mmb_logging_init(in_opts->logging_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the logging infrastructure.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_init_providers(in_opts->memory_opts->providers_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the memory provider libraries.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_init_strategies(in_opts->memory_opts->strategies_opts);
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize the memory allocation strategies.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_manager_create(&g_mmb_instance.manager);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to create memory manager\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_alloc_initialize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to initialize mmbAllocationFactory.\n");
    goto UNLOCK_BAILOUT;
  }

#ifdef HAVE_LOOP_ANALYSIS
  stat = mmb_loop_initialize();
  if(stat != MMB_OK) {
    MMB_ERR("Failed to initialize mmb_loop\n");
    goto UNLOCK_BAILOUT;
  } 
#endif 

#ifdef HAVE_COST_MODEL
  stat = mmb_cost_model_initialize(&g_mmb_instance.cm_handle);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to initialize cost model library\n");
    goto UNLOCK_BAILOUT;
  } 
#endif

  g_mmb_instance.is_initialized = true;
  MMB_DEBUG("Mamba library initialised, version: %s\n", mmb_version());

UNLOCK_BAILOUT:

  if (!_in_opts)
    mmb_options_destroy(in_opts);

  mmb_state_unlock();

  return stat;
}

mmbError mmb_is_initialized(bool *initialized) {
  if(!initialized)
    return MMB_INVALID_ARG;
  *initialized = g_mmb_instance.is_initialized;
  return MMB_OK;
}

mmbError mmb_register_memory(const mmbMemLayer layer,
                             const mmbExecutionContext ex_context,
                             const mmbMemSpaceConfig *config_opts,
                             mmbMemSpace **new_space)
{
  const mmbMemSpaceConfig default_config = MMB_MEMSPACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  /* /!\ mmbMemSpaceConfig->mem_size is given in bytes, not MiB. */
  mmbMemSpaceConfig space_config = *config_opts;
  if (SIZE_MAX == space_config.size_opts.mem_size // SIZE_MAX is a special value
      || SIZE_MAX / (1024*1024) < space_config.size_opts.mem_size)
    space_config.size_opts.mem_size = SIZE_MAX;
  else
    space_config.size_opts.mem_size *= 1024 * 1024;
  return mmb_manager_register_memory(g_mmb_instance.manager, layer, ex_context, &space_config,
                                     new_space);
}

mmbError mmb_request_space(const mmbMemLayer layer,
                           const mmbExecutionContext ex_con,
                           const mmbMemSpaceConfig *config_opts,
                           mmbMemSpace **out_space)
{
  const mmbMemSpaceConfig default_config = MMB_MEMSPACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  return mmb_manager_request_space(g_mmb_instance.manager, layer, ex_con, config_opts, out_space);
}

mmbError mmb_create_interface(mmbMemSpace *space,
                              const mmbMemInterfaceConfig *config_opts,
                              mmbMemInterface **out_interface)
{
  mmbMemInterfaceConfig default_config = MMB_MEMINTERFACE_CONFIG_DEFAULT;
  default_config.name = "";
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  return mmb_memspace_create_interface(space, config_opts, out_interface);
}

mmbError mmb_request_interface(mmbMemSpace *space,
                               const mmbMemInterfaceConfig *config_opts,
                               mmbMemInterface **out_interface)
{
  const mmbMemInterfaceConfig default_config = MMB_MEMINTERFACE_CONFIG_DEFAULT;
  if (NULL == config_opts) {
    config_opts = &default_config;
  }
  return mmb_memspace_request_interface(space, config_opts, out_interface);
}

mmbError mmb_destroy_interface(mmbMemInterface *interface)
{
  return mmb_memspace_destroy_interface(interface);
}

mmbError mmb_memspace_configure(const mmbMemSpaceConfig *config_opts,
                                mmbMemSpace *space)
{
  if (NULL == config_opts) {
    MMB_WARN("NULL pointer given for configure option, no change to be done.\n");
    return MMB_OK;
  }
  return mmb_memspace_configure_defaults(g_mmb_instance.manager, config_opts, space);
}

mmbError mmb_dump_memory_state(FILE *out_stream)
{
  return mmb_manager_report_state(g_mmb_instance.manager, out_stream);
}

mmbError mmb_get_hwloc_topology(mmb_hwloc_topology_t **topo) {
  return mmb_manager_get_topology(g_mmb_instance.manager, topo);
}

mmbError mmb_discovery_is_enabled(int *is_enabled)
{
  return mmb_manager_discovery_is_enabled(is_enabled);
}

mmbError mmb_statistics_get_current_statistics(mmbMemInterface *interface,
                                               size_t *n_allocations,
                                               size_t *bytes_allocated)
{
  if(!g_mmb_instance.is_initialized) {
    MMB_ERR("Mamba library not initialized\n");
    return MMB_ERROR;
  }
  return mmb_manager_get_interface_current_statistics(g_mmb_instance.manager,
                                                      interface, n_allocations,
                                                      bytes_allocated);
}

mmbError mmb_statistics_get_total_statistics(mmbMemInterface *interface,
                                             size_t *n_allocations,
                                             size_t *bytes_allocated)
{
  if(!g_mmb_instance.is_initialized) {
    MMB_ERR("Mamba library not initialized\n");
    return MMB_ERROR;
  }
  return mmb_manager_get_interface_total_statistics(g_mmb_instance.manager,
                                                    interface, n_allocations,
                                                    bytes_allocated);
}

mmbError mmb_statistics_get_record_count(size_t *count)
{
  if(!g_mmb_instance.is_initialized) {
    MMB_ERR("Mamba library not initialized\n");
    return MMB_ERROR;
  }
  return mmb_manager_get_stats_record_count(g_mmb_instance.manager, count);
}

#if HAVE_COST_MODEL
mmbError mmb_report_cost_model_state() {
 return mmb_cost_model_write_hw_devices(g_mmb_instance.cm_handle);
}
#endif

mmbError mmb_finalize(void) {
  mmbError stat = MMB_OK;

  mmb_state_lock();

  if(!g_mmb_instance.is_initialized) {
      MMB_ERR("Mamba library not initialized\n");
      stat = MMB_ERROR;
      goto UNLOCK_BAILOUT;
  }

#ifdef HAVE_COST_MODEL
  stat = mmb_cost_model_finalize(g_mmb_instance.cm_handle);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to finalize cost model library\n");
    goto UNLOCK_BAILOUT;
  } 
  g_mmb_instance.cm_handle = NULL;
#endif

  stat = mmb_manager_destroy(g_mmb_instance.manager);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to destroy memory manager\n");
    goto UNLOCK_BAILOUT;
  }
  g_mmb_instance.manager = NULL;

  stat = mmb_alloc_finalize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the mmbAllocationFactory.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_finalize_providers();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the memory provider libraries.\n");
    goto UNLOCK_BAILOUT;
  }

  stat = mmb_finalize_strategies();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the memory allocation strategies.\n");
    goto UNLOCK_BAILOUT;
  }

#ifdef HAVE_LOOP_ANALYSIS
  stat = mmb_loop_finalize();
  if(stat != MMB_OK) {
    MMB_ERR("Failed to finalize mmb_loop\n");
    goto UNLOCK_BAILOUT;
  } 
#endif

  stat = mmb_logging_finalize();
  if (MMB_OK != stat) {
    MMB_ERR("Failed to finalize the logging infrastructure.\n");
    goto UNLOCK_BAILOUT;
  }

  g_mmb_instance.is_initialized = false;

UNLOCK_BAILOUT:
  mmb_state_unlock();

 return stat;
}

/* protecting array state */
static inline void mmb_tiling_rdlock(mmbArray *mba) {
  int stat = pthread_rwlock_rdlock(&mba->tiling_lock);
  assert(stat==0);
}

static inline void mmb_tiling_wrlock(mmbArray *mba) {
  int stat = pthread_rwlock_wrlock(&mba->tiling_lock);
  assert(stat==0);
}

static inline void mmb_tiling_unlock(mmbArray *mba) {
  int stat = pthread_rwlock_unlock(&mba->tiling_lock);
  assert(stat==0);
}

mmbError mmb_tile_count(mmbArray *in_mba, int *out_count) {
  mmbError stat = MMB_OK;
  if(!in_mba || !out_count) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  /* Return counted tiles */
  *out_count = in_mba->tiling->total;
BAILOUT:
  return stat;
}

mmbError mmb_tiling_dimensions(mmbArray *in_mba, mmbDimensions **out_dims) {
  mmbError stat = MMB_OK;

  if(!in_mba || !out_dims) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  if(!in_mba->tiling) {
    MMB_ERR("mmbArray is not tiled\n");
    return MMB_ERROR;
  }

  mmbDimensions* dims;
  stat =  mmb_dimensions_create(0, &dims);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to create mmb dimensions object\n");
    goto BAILOUT;
  } 

  stat = mmb_dimensions_copy(dims, &in_mba->tiling->tile_count);
  if(stat != MMB_OK) {
    MMB_ERR("Could not retrieve mmbDimensions from array tiling\n");
    goto BAILOUT;
  }

  *out_dims = dims;

BAILOUT:
  return stat;
}

mmbError mmb_tile_at(mmbArray *in_mba, mmbIndex *in_idx,
                     mmbArrayTile **out_tile) {
  mmbError stat = MMB_OK;

  if(!in_mba || !in_idx || !out_tile) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  *out_tile = NULL;
  if(!in_mba->tiling) {
    stat = MMB_ERROR;
    MMB_ERR("mmbArray not tiled\n");
    goto BAILOUT;
  }

  /* Lookup in hashmap for tile at in_idx */
  size_t tile_idx = 0;
  stat = nd_to_1d_idx(in_idx, &in_mba->tiling->tile_count, &tile_idx);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to convert nd index to 1d index\n");
    goto BAILOUT;
  }

  mmbArrayTile *t = NULL;
  mmb_tiling_rdlock(in_mba);
  mmb_hash_iter_t k = mmb_hash_get_array_tiles(in_mba->tiling->tiles, tile_idx);
  if(k != mmb_hash_end_array_tiles(in_mba->tiling->tiles)) {
    t = mmb_hash_value_array_tiles(in_mba->tiling->tiles, k);
  /* If found, move to top tile of the stack and return pointer */
    while (t->next != NULL){
      t = t->next;
    }
    *out_tile = t;
    stat = MMB_OK;
    goto BAILOUT_UNLOCK;
  }
  
  mmb_tiling_unlock(in_mba);

  /* Otherwise, create tile and return using default memory space */ 

  /* Typical threaded usage would be multiple threads asking for different
    tiles, so we let each thread make its own tile before insertion
    This results in some wasted cycles if many threads ask for the same 
    not-yet-constructed tile, but minimises blocking in the more-common scenario
    where they ask for different not-yet-constructed tiles. */
  t = (mmbArrayTile *)calloc(1, sizeof(mmbArrayTile));
  if(!t) {
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }

  size_t *dims = malloc(sizeof(size_t[6][in_mba->layout->n_dims]));
  if (NULL == dims) {
    free(t);
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  t->lower   = &dims[0 * in_mba->layout->n_dims];
  t->upper   = &dims[1 * in_mba->layout->n_dims];
  t->alower  = &dims[2 * in_mba->layout->n_dims];
  t->aupper  = &dims[3 * in_mba->layout->n_dims];
  t->dim     = &dims[4 * in_mba->layout->n_dims];
  t->abs_dim = &dims[5 * in_mba->layout->n_dims];

  t->interface = in_mba->interface;
  t->access = in_mba->access;
  /* TODO: Could just use in_mba->layout pointer instead of duplicating? */
  stat = mmb_layout_create_copy(in_mba->layout, &t->layout);
  if (stat != MMB_OK) {
    MMB_ERR("Failed to copy-construct mmbLayout\n");
    free(t);
    free(dims);
    goto BAILOUT;
  }  

  for (unsigned i = 0; i < t->layout->n_dims; i++) {
    t->abs_dim[i] = in_mba->dims.d[i];
    t->lower[i] = in_idx->d[i] * in_mba->tiling->tile_size.d[i];
    t->upper[i] = (in_idx->d[i] + 1) * in_mba->tiling->tile_size.d[i];
    if (t->upper[i] > t->abs_dim[i])
      t->upper[i] = t->abs_dim[i];
    t->alower[i] = t->lower[i];
    t->aupper[i] = t->upper[i];
    t->dim[i] = t->upper[i] - t->lower[i];
  }

  /* Check contiguity */
  /* TODO: This only checks if tile is 1d, or if it covers whole array, 
   * not if subtiles are contiguous */
  int contiguous = 1;
  if(t->layout->n_dims > 1) {
    for (unsigned i = 0; i < t->layout->n_dims; i++) {
      if(t->abs_dim[i] != t->dim[i]) {
        contiguous = 0;
      }
    }
  }

  if(contiguous) {
    t->access |= MMB_CONTIGUOUS;
  }

  /* This pointer is offset using indexing macros */
  t->allocation = in_mba->allocation;

  /* Set tile index for hashmap */
  t->tidx = tile_idx;

  /* Initialise the tile lock */
  pthread_mutex_init(&t->lock, NULL); 
  
  mmb_tiling_wrlock(in_mba);
  /* Check if someone else already created this tile in the meantime
   * if so, free local data and return pre-existing tile */
  mmbArrayTile *t_temp;

  k = mmb_hash_get_array_tiles(in_mba->tiling->tiles, tile_idx);
  if(k != mmb_hash_end_array_tiles(in_mba->tiling->tiles)) {
    t_temp = mmb_hash_value_array_tiles(in_mba->tiling->tiles, k);
    /* We dont need the tile we constructed ourselves anymore */
    stat = mmb_layout_destroy(t->layout);
    if(stat != MMB_OK)
      MMB_WARN("Failed to destroy temporary layout object\n");
    free(dims);
    pthread_mutex_destroy(&t->lock); 
    free(t);

    t = t_temp;
    /* Move to top tile of the stack and return pointer */
    while (t->next != NULL){
      t = t->next;
    }
    *out_tile = t;
    stat = MMB_OK;
    goto BAILOUT_UNLOCK;
  }
  int ret;
  k = mmb_hash_put_array_tiles(in_mba->tiling->tiles, tile_idx, &ret);
  if(ret<0) {
    MMB_ERR("Unable to insert tile into hash map\n");
    stat = MMB_ERROR;
    goto BAILOUT_UNLOCK;
  }
  mmb_hash_value_array_tiles(in_mba->tiling->tiles, k) = t;

  *out_tile = t;
BAILOUT_UNLOCK:
  mmb_tiling_unlock(in_mba);
BAILOUT:
  return stat;
}

mmbError mmb_tile_at_1d(mmbArray *in_mba, size_t in_idx,
                     mmbArrayTile **out_tile){
  mmbIndex idx = {1, &in_idx};
  return mmb_tile_at(in_mba, &idx, out_tile);
}

mmbError mmb_tile_at_2d(mmbArray *in_mba, size_t in_idx0, size_t in_idx1,
                     mmbArrayTile **out_tile){
  size_t idx_arr[2] = {in_idx0, in_idx1};
  mmbIndex idx = {2, &idx_arr[0]};
  return mmb_tile_at(in_mba, &idx, out_tile);

}

mmbError mmb_tile_at_3d(mmbArray *in_mba, size_t in_idx0, size_t in_idx1,
                        size_t in_idx2, mmbArrayTile **out_tile){
  size_t idx_arr[3] = {in_idx0, in_idx1, in_idx2};
  mmbIndex idx = {3, &idx_arr[0]};
  return mmb_tile_at(in_mba, &idx, out_tile);

}


mmbError mmb_tile_duplicate_at(mmbArray *in_mba, mmbIndex *in_idx,
                               mmbMemInterface *in_interface,
                               mmbArrayTile **out_tile)
{
  mmbError stat = MMB_OK;
  mmbArrayTile *t = NULL;
  stat = mmb_tile_at(in_mba, in_idx, &t);
  if (MMB_OK != stat) {
    MMB_ERR("Unable to retrive the corresponding tile.\n");
    return stat;
  }
  return mmb_tile_duplicate(t, in_interface, in_mba->access, in_mba->layout, out_tile);
}

