/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>

#include "mmb_tile.h"
#include "mmb_logging.h"
#include "mmb_memory.h"
#include "mamba.h"

#include "i_memory.h"


static inline void mmb_tile_lock(mmbArrayTile *t) {
  int stat = pthread_mutex_lock(&t->lock);
  assert(stat==0);
}
static inline void mmb_tile_unlock(mmbArrayTile *t) {
  int stat = pthread_mutex_unlock(&t->lock);
  assert(0 == stat);
}

mmbError mmb_tile_alloc_data(mmbArrayTile *in_tile) {
  /* Validate args */
  if(in_tile == NULL)
    return MMB_ERROR;

  if(in_tile->allocation != NULL)
    return MMB_ERROR;

  mmbError stat = MMB_OK;
  size_t total_bytes;

  mmbDimensions dims = mmb_dimensions_empty;
  stat = mmb_dimensions_fill(&dims, in_tile->dim, in_tile->layout->n_dims);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to create tile dimensions for allocation\n");
    goto BAILOUT;
  }
  stat = mmb_layout_buffer_size(in_tile->layout, &dims, &total_bytes);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to calculate tile buffer size\n");
    goto BAILOUT;
  }
  stat = mmb_dimensions_resize(&dims, 0);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to destroy dimensions object\n");
    goto BAILOUT;
  }

  /* TODO: We need to consider access type here too */
  stat = mmb_allocate(total_bytes, in_tile->interface, &in_tile->allocation);
  if(stat != MMB_OK) {
    MMB_ERR("Unable to allocate %zu bytes for tile\n", total_bytes);
    goto BAILOUT;
  }

  in_tile->allocation_owned = 1;

BAILOUT:
  return stat;
}

mmbError mmb_tile_pack(mmbArrayTile *tile,
                       mmbAllocation *host_alloc,
                       mmbAllocation *device_alloc)
{
  mmbError stat = MMB_OK;
  if (NULL == tile) {
    MMB_ERR("Invalid tile cannot be packed.\n");
    return MMB_INVALID_ARG;
  }
  if (NULL == host_alloc || NULL == host_alloc->ptr) {
    MMB_ERR("Invalid host located allocation provided cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  if (NULL == device_alloc || NULL == device_alloc->ptr) {
    MMB_ERR("Invalid device located allocation provided cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  /* Write Tile */
  mmbArrayTile *buffer = MMBALLOC2TBUF(host_alloc, mmbArrayTile *);
  memcpy(buffer, tile, sizeof(mmbArrayTile));
  char *device_tmp = MMBALLOC2TBUF(device_alloc, char *);
  /* leave a gap just after the tile for the allocation */
  device_tmp += sizeof(mmbArrayTile);
  buffer->allocation = (mmbAllocation *)device_tmp;
  device_tmp += sizeof(mmbAllocation);
  /* write dims */
  size_t *dims = (size_t *)device_tmp;
  buffer->lower   = &dims[0 * tile->layout->n_dims];
  buffer->upper   = &dims[1 * tile->layout->n_dims];
  buffer->alower  = &dims[2 * tile->layout->n_dims];
  buffer->aupper  = &dims[3 * tile->layout->n_dims];
  buffer->dim     = &dims[4 * tile->layout->n_dims];
  buffer->abs_dim = &dims[5 * tile->layout->n_dims];
  char *tmp_buf = (char *)(buffer + 1) + sizeof(mmbAllocation);
  memcpy(tmp_buf, tile->lower, sizeof(size_t[6][tile->layout->n_dims]));
  /* write layout */
  device_tmp += sizeof(size_t[6][tile->layout->n_dims]);
  tmp_buf += sizeof(size_t[6][tile->layout->n_dims]);
  buffer->layout = (mmbLayout *)device_tmp;
  stat = mmb_layout_pack(tmp_buf, tile->layout, device_tmp);
  if (MMB_OK != stat) {
    MMB_ERR("Unable to pack layout for device tile.\n");
    return stat;
  }
  /* The allocation part that has been left out, is filled by the calling
   * function. */
  return stat;
}

/* This can be lockfree due to the tile_duplicate implementation only
 * duplicating one-to-one tiles, when we extend to one-to-many this will
 * need reviewing */
mmbError mmb_tile_copy(mmbArrayTile *dst_tile, mmbArrayTile *src_tile) {
  /* Validate args */
  if(dst_tile == NULL || src_tile == NULL){
    MMB_ERR("Arg can not be NULL\n");
    return MMB_INVALID_ARG;
  }
  if(dst_tile == src_tile){
    MMB_ERR("dst_tile==src_tile, cannot copy tile to itself\n");
    return MMB_INVALID_ARG;
  }
  mmbError stat = MMB_OK;
  /* dst_tile must be writeable */
  if(!(dst_tile->access & MMB_WRITE)) {
    MMB_ERR("Destination tile not writeable for copy\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  /* Only matching layouts are supported currently */
  int diff;
  stat = mmb_layout_cmp(src_tile->layout, dst_tile->layout, &diff);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to compare layouts\n");
    goto BAILOUT;
  }

  if(diff){
    MMB_ERR("Layouts do not match - only identical layouts currently supported for copy\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }
  /* Only contiguous tiles are supported currently */
  if(!(src_tile->access & MMB_CONTIGUOUS) || !(dst_tile->access & MMB_CONTIGUOUS)) {
    MMB_ERR("Tile layouts not contiguous - only contiguous layouts currently supported for copy\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  /* Work out offset in src and dst allocations 
    
    Compare tile dim vs tile abs dim

    if matching, no offset
    if no match, work out starting offset
  */
  bool dst_tile_alloc_is_total_alloc = true;
  for(unsigned i = 0; i < dst_tile->layout->n_dims; i++) {
    if(dst_tile->abs_dim[i] != dst_tile->dim[i]) {
      dst_tile_alloc_is_total_alloc = false;
      break;
    }
  }
  size_t dst_offset_bytes = 0;
  size_t dst_size_bytes = 0;
  if(!dst_tile_alloc_is_total_alloc) {
    /* Work out dst tile allocation offset */
    stat = mmb_tile_lower_byte_offset(dst_tile, &dst_offset_bytes);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to determine byte offset for index into destination tile\n");
      goto BAILOUT;
    }
    /* Work out dst tile buffer size */
    mmbDimensions dims = mmb_dimensions_empty;
    stat = mmb_dimensions_fill(&dims, dst_tile->dim, dst_tile->layout->n_dims);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to copy tile dimensions for allocation\n");
      goto BAILOUT;
    }
    stat = mmb_layout_buffer_size(dst_tile->layout, &dims, &dst_size_bytes);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to calculate tile buffer size\n");
      goto BAILOUT;
    }
    stat = mmb_dimensions_resize(&dims, 0);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to empty mmbDimensions\n");
      goto BAILOUT;
    }
  }
  /* Create suballocation for tile if necessary */
  mmbAllocation *dst_allocation = NULL;
  if(!dst_tile_alloc_is_total_alloc) {
    stat = mmb_allocation_get_slice(dst_tile->allocation,
                                    dst_offset_bytes, dst_size_bytes,
                                    &dst_allocation);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to get slice of \n");
      goto BAILOUT;
    }
  } else {
    dst_allocation = dst_tile->allocation;
  }

  bool src_tile_alloc_is_total_alloc = true;
  for(unsigned i = 0; i < src_tile->layout->n_dims; i++) {
    if(src_tile->abs_dim[i] != src_tile->dim[i]) {
      src_tile_alloc_is_total_alloc = false;
      break;
    }
  }
  size_t src_offset_bytes = 0;
  size_t src_size_bytes = 0;
  if(!src_tile_alloc_is_total_alloc) {
    /* Work out src tile allocation offset */
    stat = mmb_tile_lower_byte_offset(src_tile, &src_offset_bytes);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to determine byte offset for index into source tile\n");
      goto BAILOUT;
    }
    /* Work out src tile buffer size */
    mmbDimensions dims = mmb_dimensions_empty;
    stat = mmb_dimensions_fill(&dims, src_tile->dim, src_tile->layout->n_dims);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to copy tile dimensions for allocation\n");
      goto BAILOUT;
    }
    stat = mmb_layout_buffer_size(src_tile->layout, &dims, &src_size_bytes);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to calculate tile buffer size\n");
      goto BAILOUT;
    }
    stat = mmb_dimensions_resize(&dims, 0);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to empty mmbDimensions\n");
      goto BAILOUT;
    }
  }
  /* Create suballocation for tile if necessary */
  mmbAllocation *src_allocation = NULL;
  if(!src_tile_alloc_is_total_alloc) {
    stat = mmb_allocation_get_slice(src_tile->allocation,
                                    src_offset_bytes, src_size_bytes,
                                    &src_allocation);
    if(stat != MMB_OK) {
      MMB_ERR("Unable to get slice of \n");
      goto BAILOUT;
    }
  } else {
    src_allocation = src_tile->allocation;
  }

  /* Do the copy */
  stat = mmb_copy(dst_allocation, src_allocation);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to copy src to dst tile\n");
    goto BAILOUT;
  }

  /* Copy the metadata to have a coherent tile on the device if necessary */
  mmbExecutionContext dst_ex_con;
  stat = mmb_allocation_get_execution_context(dst_allocation, &dst_ex_con);
  if (MMB_OK != stat) {
    MMB_ERR("Unable to read destination tiling execution context.\n");
    goto BAILOUT;
  }
  if (MMB_CPU != dst_ex_con && NULL == src_tile->space_local_metadata) {
    mmbAllocation *src_alloc, *dst_alloc;
    size_t metadata_size = 0;
    /* Prepare the buffer to send to the device */
    stat = mmb_layout_get_sizeof(dst_tile->layout, &metadata_size);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to determine the size of the metadata to transfert to the "
          "device.\n");
      goto BAILOUT;
    }
    metadata_size += sizeof(mmbAllocation);
    metadata_size += sizeof(mmbArrayTile);
    metadata_size += sizeof(size_t[6][dst_tile->layout->n_dims]);
    stat = mmb_allocate(metadata_size, src_tile->allocation->interface, &src_alloc);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to allocate memory for CPU side buffer for tile metadata.\n");
      return MMB_OUT_OF_MEMORY;
    }
    stat = mmb_allocate(metadata_size, dst_tile->allocation->interface, &dst_alloc);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to allocate memory for device side tile metadata.\n");
      mmb_free(src_alloc);
      return stat;
    }
    stat = mmb_tile_pack(dst_tile, src_alloc, dst_alloc);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to pack metadata for device side tile.\n");
      mmb_free(src_alloc);
      mmb_free(dst_alloc);
      return stat;
    }
    /* Update de mmbAllocation region of the buffer */
    mmbAllocation *device_alloc;
    device_alloc = (mmbAllocation *)(MMBALLOC2TBUF(src_alloc, mmbArrayTile *) + 1);
    memcpy(device_alloc, dst_allocation, sizeof(mmbAllocation));
    /* Transfert the tile metadata to the device */
    stat = mmb_copy(dst_alloc, src_alloc);
    if (MMB_OK != stat) {
      MMB_ERR("Unable to transfert the metadata to the device.\n");
      mmb_free(src_alloc);
      mmb_free(dst_alloc);
      return stat;
    }
    mmb_free(src_alloc);
    dst_tile->space_local_metadata = dst_alloc;
  }


  /* Destroy the allocations if necessary */
  if(!dst_tile_alloc_is_total_alloc)
    mmb_allocation_destroy(dst_allocation);
  if(!src_tile_alloc_is_total_alloc)
    mmb_allocation_destroy(src_allocation);

BAILOUT:
  return stat;
}

mmbError mmb_tile_duplicate(mmbArrayTile *in_tile, mmbMemInterface *in_interface,
                            mmbAccessType in_access, mmbLayout *in_layout,  
                            mmbArrayTile **out_tile) {

  mmbError stat = MMB_OK;
  /* 
    Validate input arguments
  */
  if(!in_tile || !in_interface || !in_layout || !out_tile) {
    MMB_ERR("Arg can not be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  /* If in_tile is read-only, and in_access is writeable, return error */
  if(!(in_tile->access & MMB_WRITE) && (in_access & MMB_WRITE)) {
    MMB_ERR("Unable to duplicate read-only tile to writeable tile.\n");
    return MMB_ERROR;
  }

  /* Lock tile for access. Currently you can only duplicate the top tile in the 
   * stack, so this locking can be simple. Eventually we will support duplicating 
   * many read-only tiles from a single source tile simultaneously, in which case this
   * will require updating */
  mmb_tile_lock(in_tile);

  /* Can only duplicate top tile in the stack */
  if(in_tile->next != NULL){
    MMB_ERR("Tile for duplication must not have dependent tiles\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  /* Initialise new tile from input parameters */
  mmbArrayTile *t = (mmbArrayTile *)calloc(1, sizeof(mmbArrayTile));
  if (NULL == t) {
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }


  size_t *dims = malloc(sizeof(size_t[6][in_layout->n_dims]));
  if (NULL == dims) {
    free(t);
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  t->lower   = &dims[0 * in_layout->n_dims];
  t->upper   = &dims[1 * in_layout->n_dims];
  t->alower  = &dims[2 * in_layout->n_dims];
  t->aupper  = &dims[3 * in_layout->n_dims];
  t->dim     = &dims[4 * in_layout->n_dims];
  t->abs_dim = &dims[5 * in_layout->n_dims];

  t->interface = in_interface;
  t->access = in_access;
  stat = mmb_layout_create_copy(in_layout, &t->layout);
  if (stat != MMB_OK) {
    free(t);
    free(dims);
    goto BAILOUT;
  }

  /* The duplicated tile lower and upper are now 0:dim in each dimension */
  for (size_t i = 0; i < in_layout->n_dims; i++) {
    t->abs_dim[i] = in_tile->dim[i];
    t->lower[i] = 0;
    t->upper[i] = in_tile->dim[i];
    t->alower[i] = in_tile->lower[i];
    t->aupper[i] = in_tile->upper[i];
    t->dim[i] = in_tile->dim[i];
  }
  t->tidx = in_tile->tidx;

  /* Allocate memory for new tile */
  stat = mmb_tile_alloc_data(t);
  if (stat != MMB_OK) {
    free(t);
    free(dims);
    goto BAILOUT;
  } 

  /* Duplicated tile is now contiguous */
  t->access |= MMB_CONTIGUOUS;

  /* Copy input tile to new tile */
  stat = mmb_tile_copy(t, in_tile);
  if (stat != MMB_OK) {
    free(t);
    free(dims);
    goto BAILOUT;
  } 

  /* Initialise the tile lock */
  pthread_mutex_init(&t->lock, NULL); 

  /* Add to the stack of tiles */
  in_tile->next = t;
  t->prev = in_tile;

  /* If access type is writeable, then lock the tile below from being written */ 
  if(in_access & MMB_WRITE){
    in_tile->access |= MMB_WRITE_LOCKED;
  }

  /* Return tile */
  *out_tile = t;
  
BAILOUT:
  mmb_tile_unlock(in_tile);
  return stat;
}

mmbError mmb_tile_get_n_dims(mmbArrayTile *in_tile,
                             size_t *n_dims){

  mmbError stat = MMB_OK;
  /* Check parameters */
  if (NULL == in_tile) {
    MMB_ERR("Invalid tile given as parameter.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL == n_dims) {
    MMB_ERR("Invalid n_dims pointer given as parameter.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }

  *n_dims = in_tile->layout->n_dims;

BAILOUT:
  return stat;
}

mmbError mmb_tile_get_space_local_handle(mmbArrayTile *in_tile,
                                         mmbArrayTile **out_tile)
{
  mmbError stat = MMB_OK;
  /* Check parameters */
  if (NULL == in_tile) {
    MMB_ERR("Invalid tile given as parameter.\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL == out_tile) {
    MMB_ERR("Invalid handle given as parameter. \"out_tile\" cannot be NULL\n");
    stat = MMB_INVALID_ARG;
    goto BAILOUT;
  }
  if (NULL != in_tile->space_local_metadata) {
    *out_tile = MMBALLOC2TBUF(in_tile->space_local_metadata, mmbArrayTile *);
  } else {
    *out_tile = in_tile;
  }
BAILOUT:
  return stat;
}

mmbError mmb_tile_migrate(mmbArrayTile *in_tile, mmbMemInterface *in_interface,
                          mmbAccessType in_access, mmbLayout *in_layout) {

  /*
    Validate input arguments

    If in_tile is read-only, and in_access is writeable
      Check if tile below is writeable
      If not, return error.

    Allocate new tile with interface / layout / access type

    Copy data to new tile

    Replace old tile with new in tile stack

    Free old tile
  */
  (void) in_tile; (void) in_interface; (void) in_access; (void) in_layout;
  mmbError stat = MMB_UNIMPLEMENTED;
  return stat;
}

mmbError mmb_tile_merge(mmbArrayTile *in_tile, mmbMergeType in_merge) {

  /* 
      Validate input arguments

      switch on in_merge
        case OVERWRITE:
          get reference to tile_below
          copy in_tile to tile_below
      
      Other cases we will consider:
        pull data from tile_below to in_tile
        fancy merges (e.g modified entries only, halos only, user defined function)

  */

  mmbError stat = MMB_OK;
  if(!in_tile) {
    stat = MMB_ERROR;
    goto BAILOUT;
  }
  mmb_tile_lock(in_tile);
  /* If tile is bottom of stack, or if it was read only, free the tile. */
  if(in_tile->prev == NULL || !(in_tile->access & MMB_WRITE)) {
    mmb_tile_unlock(in_tile);
    stat = mmb_tile_destroy(in_tile);
    if(stat != MMB_OK) {
      MMB_ERR("Failed to destroy tile during no copy merge\n");
    }
    goto BAILOUT;
  }

  mmb_tile_lock(in_tile->prev);
  switch(in_merge) {
    case MMB_DISCARD: /* no action needed */
      break;
    case MMB_OVERWRITE:
    {
      mmbArrayTile *prev = in_tile->prev;
      if(!(prev->access & MMB_WRITE)) {
        MMB_ERR("Trying to merge to read-only tile\n");
        stat = MMB_ERROR;
        goto BAILOUT_UNLOCK_FREE;
      }

      if(!(prev->access & MMB_CONTIGUOUS) || !(in_tile->access & MMB_CONTIGUOUS)) {
        MMB_ERR("Only contiguous tile merge currently supported\n");
        stat = MMB_ERROR;
        goto BAILOUT_UNLOCK_FREE;
      }

      stat = mmb_tile_copy(prev, in_tile);
      if(stat != MMB_OK) {
        MMB_ERR("Failed to copy tile data during merge\n");
        goto BAILOUT_UNLOCK_FREE;
      }
    }
    break;
    default:
      MMB_ERR("Unsupported merge type\n");
      stat = MMB_ERROR;
      goto BAILOUT_UNLOCK_FREE;
    break;
  }

  in_tile->prev->next = NULL;

BAILOUT_UNLOCK_FREE:
  mmb_tile_unlock(in_tile->prev);
  mmb_tile_unlock(in_tile);

  /* Only destroy the tile if merge was successful */
  if(stat == MMB_OK){
    stat = mmb_tile_destroy(in_tile);
    if(stat != MMB_OK) {
      MMB_ERR("Failed to destroy merged tile\n");
      goto BAILOUT;
    }    
  }

BAILOUT:
  return stat;
}

mmbError mmb_tile_destroy(mmbArrayTile *in_tile) {

  /* 
    Validate input argument

    If tile has other tiles above it (in_tile->next != null)
      return error
    
    If tile has other tiles below it (in_tile->prev != null)
      Clear next ptr from lower tile and if it was write-locked, unlock it.

    if tile data is owned
      free tile data (opposite of allocation in duplicate/migrate)
  */

  mmbError stat = MMB_OK;
  if(!in_tile) {
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  if(in_tile->next != NULL) {
    MMB_ERR("Cannot free tile with dependent tiles\n");
    stat = MMB_ERROR;
    goto BAILOUT;
  }

  if(in_tile->prev != NULL) {
    mmbArrayTile *prev = in_tile->prev;
    prev->next = NULL;
    if(prev->access & MMB_WRITE_LOCKED) {
      prev->access &= !MMB_WRITE_LOCKED;
    }
  }

  if(in_tile->allocation_owned) {
    stat = mmb_free(in_tile->allocation);
    if(stat != MMB_OK) {
      MMB_ERR("Failed to free tile allocation\n");
      goto BAILOUT;
    }
  }

  if (NULL != in_tile->space_local_metadata) {
    stat = mmb_free(in_tile->space_local_metadata);
    if (MMB_OK != stat) {
      MMB_ERR("Failed to free device tiling metadata.\n");
      goto BAILOUT;
    }
  }

  stat = mmb_layout_destroy(in_tile->layout);
  if(stat != MMB_OK) {
    MMB_ERR("Failed to free tile layout\n");
    goto BAILOUT;
  }

  /* Initialise the tile lock */
  pthread_mutex_destroy(&in_tile->lock); 

  free(in_tile->lower);
  free(in_tile);

  MMB_DEBUG("Tile destroyed\n");

BAILOUT:
  return stat;
}



mmbError mmb_tile_lower_byte_offset(mmbArrayTile *tile, size_t *n_bytes) {
  mmbError stat = MMB_OK;
  if(!tile) {
    stat = MMB_ERROR;
    goto BAILOUT;
  }
  /* Ptr comparison between mmbAllocation and tile starting index */
  /* TODO: This should support ND */
  /* TODO: Support padding here? */
  unsigned char* tile_lower_ptr = NULL;
  switch(tile->layout->n_dims) {
    case 1:
      tile_lower_ptr = (unsigned char*)&(MMB_IDX_1D(tile, tile->lower[0], unsigned char));
    break;
    case 2:
      tile_lower_ptr = (unsigned char*)&(MMB_IDX_2D(tile, tile->lower[0], tile->lower[1], unsigned char));
    break;
    default:
      MMB_ERR("Unsupported dimensions for tile indexing\n");
      stat = MMB_ERROR;
      goto BAILOUT;
    break;
  }
  size_t n_elements = tile_lower_ptr - (unsigned char*)tile->allocation->ptr;
  *n_bytes = n_elements * tile->layout->element.size_bytes;
BAILOUT:
  return stat;
}
