/*
 * Copyright (C) 2018-2020 Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#if HAVE_MEMKIND
#include <hbwmalloc.h>
#else
static inline int hbw_check_available() { return 1; }
#endif /* HAVE_MEMKIND */

#include "mmb_error.h"
#include "mmb_logging.h"
#include "mmb_memory.h"

#include "../i_memory.h"
#include "../interface.h"
#include "system.h"
#include "system_cuda.h"
#include "system_opencl.h"
#include "system_hip.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbMemInterface
 * @{ */

/**
 * @name GNU/Linux system memory provider
 * @{ */

mmbError mmb_init__system(mmbMemInterface *interface)
{
  mmbError stat = MMB_OK;
  mmbMemLayer layer = interface->space->layer;
  mmbExecutionContext ex_con = interface->space->ex_context;

  MMB_DEBUG("Calling linux initializer()\n");

  if (MMB_GDRAM == layer && MMB_GPU_OPENCL == ex_con) {
    stat = mmb_init__system_opencl(interface);
    if (MMB_OK != stat) {
      MMB_ERR("Failed to initialize OpenCL interface.\n");
      return stat;
    }
  }

  return stat;
}

mmbError mmb_finalize__system(mmbMemInterface *interface)
{
  mmbError stat = MMB_OK;
  mmbMemLayer layer = interface->space->layer;
  mmbExecutionContext ex_con = interface->space->ex_context;

  MMB_DEBUG("Calling linux finalizer()\n");

  if (MMB_GDRAM == layer && MMB_GPU_OPENCL == ex_con) {
    stat = mmb_finalize__system_opencl(interface);
    if (MMB_OK != stat) {
      MMB_ERR("Failed to finalize OpenCL interface.\n");
      return stat;
    }
  }

  return stat;
}

mmbError mmb_supports__system(const mmbMemLayer layer, bool *is_supported)
{
  if (NULL == is_supported) {
    return MMB_INVALID_ARG;
  }
  switch (layer) {
    case MMB_DRAM:
    case MMB_GDRAM:
      *is_supported = true;
      break;
    case MMB_HBM:
      *is_supported = (0 == hbw_check_available());
      break;
    case MMB_NVDIMM:
    case MMB_NADRAM:
    case MMB_SSD:
    default:
      *is_supported = false;
      break;
  }
  return MMB_OK;
}

/**
 * @name GNU/Linux malloc DRAM memory provider
 * @{ */

mmbError mmb_allocate__system_malloc(const size_t n_bytes,
                                     mmbMemInterface *interface,
                                     const mmbAllocateOptions *opts,
                                     void **out_allocation)
{
  (void) interface;
  (void) opts;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling system_malloc allocator()\n");

  void *ptr = NULL;
  switch (interface->space->layer) {
    case MMB_DRAM:
      ptr = malloc(n_bytes);
      break;
    case MMB_HBM:
#if HAVE_MEMKIND
      ptr = hbw_malloc(n_bytes);
      break;
#endif /* HAVE_MEMKIND */
    default:
      MMB_ERR("Invalid layer (%s), unavailable for system provider.\n",
              mmb_layer_get_string(interface->space->layer));
      stat = MMB_ERROR;
      goto BAILOUT;
  }
  if(NULL == ptr) {
    MMB_ERR("Could not allocate %zu bytes for mmbAllocation\n", n_bytes);
    stat = MMB_OUT_OF_MEMORY;
    goto BAILOUT;
  }
  *out_allocation = ptr;

BAILOUT:
  return stat;
}

mmbError mmb_free__system_malloc(void *allocation, mmbMemInterface *interface,
                                 const mmbAllocateOptions *opts,
                                 const size_t n_bytes)
{
  (void) opts;
  (void) n_bytes;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling system_malloc deallocator(%p)\n", allocation);

  switch (interface->space->layer) {
    case MMB_DRAM:
      free(allocation);
      break;
    case MMB_HBM:
#if HAVE_MEMKIND
      hbw_free(allocation);
      break;
#endif /* HAVE_MEMKIND */
    default:
      stat = MMB_INVALID_LAYER;
      goto BAILOUT;
  }

BAILOUT:
  return stat;
}

void* i_mmb_copy__system_malloc(void *dst, const size_t doffset,
                                const void *src, const size_t soffset,
                                const size_t width)
{
  dst = (unsigned char *) dst + doffset;
  src = (const unsigned char *) src + soffset;
  return memcpy(dst, src, width);
}

mmbError mmb_copy__system_malloc(void *dst, mmbMemInterface *dst_inter,
                                 const mmbAllocateOptions *dst_opts,
                                 const void *src, const mmbMemInterface *src_inter,
                                 const mmbAllocateOptions *src_opts,
                                 const size_t n_bytes)
{
  return mmb_copy_1d__system_malloc(dst, dst_inter, dst_opts, 0,
                                    src, src_inter, src_opts, 0,
                                    n_bytes);
}

mmbError mmb_copy_1d__system_malloc(void *dst, mmbMemInterface *dst_inter,
                                    const mmbAllocateOptions *dst_opts,
                                    const size_t doffset,
                                    const void *src,
                                    const mmbMemInterface *src_inter,
                                    const mmbAllocateOptions *src_opts,
                                    const size_t soffset,
                                    const size_t width)
{
  (void) dst_opts;
  (void) src_opts;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling system_malloc copy_1d()\n");

  if (dst_inter->space->ex_context != src_inter->space->ex_context) {
    stat = MMB_INVALID_EXECUTION_CONTEXT;
    goto BAILOUT;
  }

  MMB_DEBUG("copying %zu bytes from %p to %p.\n", width, src, dst);

  i_mmb_copy__system_malloc(dst, doffset, src, soffset, width);

BAILOUT:
  return stat;
}

mmbError mmb_copy_2d__system_malloc(void *dst, mmbMemInterface *dst_inter,
                                    const mmbAllocateOptions *dst_opts,
                                    const size_t doffset, const size_t dpitch,
                                    const void *src,
                                    const mmbMemInterface *src_inter,
                                    const mmbAllocateOptions *src_opts,
                                    const size_t soffset, const size_t spitch,
                                    const size_t width, const size_t height)
{
  (void) dst_opts;
  (void) src_opts;
  mmbError stat = MMB_OK;

  MMB_DEBUG("Calling system_malloc copy()\n");

  if (dst_inter->space->ex_context != src_inter->space->ex_context) {
    stat = MMB_INVALID_EXECUTION_CONTEXT;
    goto BAILOUT;
  }

  MMB_DEBUG("copying 2d matrix [%zu;%zu] from %p to %p.\n", width, height, src, dst);

  for (size_t row = 0; height > row; ++row)
    i_mmb_copy__system_malloc(dst, row * dpitch + doffset,
                              src, row * spitch + soffset, width);

BAILOUT:
  return stat;
}

/**  @} */

/**  @} */

/**  @} */

/**  @} */
