/*
 * Copyright (C) 2020      Cray UK
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <pthread.h>

#include "mmb_error.h"
#include "mmb_logging.h"
#include "mmb_memory.h"

#include "../i_memory.h"
#include "apply.h"
#include "i_strategy.h"
#include "i_thread_safe.h"

/**
 * @name mmb_memory
 * @{ */

/**
 * @name mmbStrategy
 * @{ */

/**
 * @name Thread Safe Strategy
 * @{ */

/**
 * @name public API
 * @{ */

mmbError mmb_strat_init_interface__thread_safe(mmbMemInterface *interface)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = interface->provider;
  unsigned char *data = interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_init_interface__apply(interface,
      mmb_strategies_va_table[MMB_THREAD_SAFE].init_interface, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].init_interface, NULL);
}

mmbError mmb_strat_finalize_interface__thread_safe(mmbMemInterface *interface)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = interface->provider;
  unsigned char *data = interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_finalize_interface__apply(interface,
      mmb_strategies_va_table[MMB_THREAD_SAFE].finalize_interface, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].finalize_interface, NULL);
}

mmbError mmb_strat_allocate__thread_safe(const size_t n_bytes,
                                         mmbMemInterface *interface,
                                         const mmbAllocateOptions *opts,
                                         mmbAllocation *allocation)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = interface->provider;
  unsigned char *data = interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_allocate__apply(n_bytes, interface, opts, allocation,
      mmb_strategies_va_table[MMB_THREAD_SAFE].allocate, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].allocate, NULL);
}

mmbError mmb_strat_free__thread_safe(mmbAllocation *allocation)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = allocation->interface->provider;
  unsigned char *data =
    allocation->interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_free__apply(allocation,
      mmb_strategies_va_table[MMB_THREAD_SAFE].free, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].free, NULL);
}

mmbError mmb_strat_copy__thread_safe(mmbAllocation *dest, const mmbAllocation *src)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = dest->interface->provider;
  unsigned char *data = dest->interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_copy__apply(dest, src,
      mmb_strategies_va_table[MMB_THREAD_SAFE].copy, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].copy, NULL);
}

mmbError mmb_strat_copy_1d__thread_safe(mmbAllocation *dst,
                                        size_t doffset,
                                        const mmbAllocation *src,
                                        size_t soffset,
                                        size_t width)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = dst->interface->provider;
  unsigned char *data = dst->interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_copy_1d__apply(
      dst, doffset, src, soffset, width,
      mmb_strategies_va_table[MMB_THREAD_SAFE].copy_1d, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].copy_1d, NULL);
}

mmbError mmb_strat_copy_2d__thread_safe(mmbAllocation *dst,
                                        size_t doffset, size_t dpitch,
                                        const mmbAllocation *src,
                                        size_t soffset, size_t spitch,
                                        size_t width, size_t height)
{
  /* data is already allocated in interface: Compute offset to strat data */
  const mmbProvider provider = dst->interface->provider;
  unsigned char *data = dst->interface->data + mmb_provider_data_sizes[provider];
  return mmb_strat_copy_2d__apply(
      dst, doffset, dpitch, src, soffset, spitch, width, height,
      mmb_strategies_va_table[MMB_THREAD_SAFE].copy_2d, data,
      mmb_strategies_va_table[MMB_STRATEGY_NONE].copy_2d, NULL);
}

/**  @} */

/** va_arg versions */

/**
 * @name private API
 * @{ */

mmbError mmb_strat_init_interface_va__thread_safe(mmbMemInterface *interface,
                                                  void *d,
                                                  va_list args)
{
  mmbError err = MMB_OK;
  int ierr;

  /** initialize specific data */
  struct mmb_strat_thread_safe_data *data = d;
  ierr = pthread_mutex_init(&data->mutex, NULL);
  if (0 != ierr) {
    MMB_ERR("Failed to initialize mutex (err=%d).\n", ierr);
    return MMB_ERROR;
  }

  /** forward call info decode */
  mmb_strat_init_va_fct fwd_init = va_arg(args, mmb_strat_init_va_fct);
  void *fwd_data = va_arg(args, void*);

  pthread_mutex_lock(&data->mutex);

  err = fwd_init(interface, fwd_data, args);

  pthread_mutex_unlock(&data->mutex);

  return err;
}

mmbError mmb_strat_finalize_interface_va__thread_safe(mmbMemInterface *interface,
                                                      void *d,
                                                      va_list args)
{
  mmbError err = MMB_OK;
  int ierr;

  /**
   * destroy specific data  but only after releasing the lock and finalizing
   * underlying structures.
   */
  struct mmb_strat_thread_safe_data *data = d;

  /** forward call info decode */
  mmb_strat_finalize_va_fct fwd_finalize = va_arg(args, mmb_strat_finalize_va_fct);
  void *fwd_data = va_arg(args, void *);

  pthread_mutex_lock(&data->mutex);

  err = fwd_finalize(interface, fwd_data, args);

  pthread_mutex_unlock(&data->mutex);

  /* finalize our data */
  ierr = pthread_mutex_destroy(&data->mutex);
  if (0 != ierr) {
    MMB_ERR("Failed to destroy mutex (err=%d).\n", ierr);
    /* in case of error, proritize the forwarding
     * of the error over sending our plain one.   */
    if (MMB_OK == err)
      err = MMB_ERROR;
  }

  return err;
}

mmbError mmb_strat_allocate_va__thread_safe(const size_t     n_bytes,
                                            mmbMemInterface *interface,
                                            const mmbAllocateOptions *opts,
                                            void            *d,
                                            void           **allocation,
                                            va_list          args)
{
  mmbError err;
  /** Check parameters */
  if (NULL == d) {
    MMB_ERR("Data parameter cannot be NULL.\n");
    return MMB_INVALID_ARG;
  }
  struct mmb_strat_thread_safe_data *data = d;

  /** forward call info decode */
  mmb_strat_allocate_va_fct fwd_alloc = va_arg(args, mmb_strat_allocate_va_fct);
  void *fwd_data = va_arg(args, void *);

  pthread_mutex_lock(&data->mutex);

  err = fwd_alloc(n_bytes, interface, opts, fwd_data, allocation, args);

  pthread_mutex_unlock(&data->mutex);

  return err;
}

mmbError mmb_strat_free_va__thread_safe(void            *allocation,
                                        void            *d,
                                        mmbMemInterface *interface,
                                        const mmbAllocateOptions *opts,
                                        const size_t     n_bytes,
                                        va_list          args)
{
  mmbError err = MMB_OK;
  /** Check parameters */
  if (NULL == d) {
    MMB_WARN("Data parameter should not be NULL.\n");
    return MMB_INVALID_ARG;
  }
  struct mmb_strat_thread_safe_data *data = d;

  /* forward call info decode */
  mmb_strat_free_va_fct fwd_free = va_arg(args, mmb_strat_free_va_fct);
  void *fwd_data = va_arg(args, void *);

  pthread_mutex_lock(&data->mutex);

  err = fwd_free(allocation, fwd_data, interface, opts, n_bytes, args);

  pthread_mutex_unlock(&data->mutex);

  return err;
}

mmbError mmb_strat_copy_va__thread_safe(void                  *dst,
                                        void                  *dst_data,
                                        mmbMemInterface       *dst_inter,
                                        const mmbAllocateOptions *dst_opts,
                                        const void            *src,
                                        const mmbMemInterface *src_inter,
                                        const mmbAllocateOptions *src_opts,
                                        const size_t           n_bytes,
                                        va_list                args)
{
  (void) dst_data;
  mmbError err = MMB_OK;
  /** Check parameters */
  if (NULL == dst_data) {
    MMB_WARN("Data parameter should not be NULL.\n");
    return MMB_INVALID_ARG;
  }
  /** forward call */
  mmb_strat_copy_va_fct fwd_copy = va_arg(args, mmb_strat_copy_va_fct);
  void *fwd_data = va_arg(args, void *);
  err = fwd_copy(dst, fwd_data, dst_inter, dst_opts, src, src_inter, src_opts, n_bytes, args);
  return err;
}

mmbError mmb_strat_copy_1d_va__thread_safe(void *dst, void *dst_data,
                                           mmbMemInterface *dst_interface,
                                           const mmbAllocateOptions *dst_opts,
                                           const size_t doffset,
                                           const void * src,
                                           const mmbMemInterface *src_interface,
                                           const mmbAllocateOptions *src_opts,
                                           const size_t soffset,
                                           const size_t width,
                                           va_list args)
{
  (void) dst_data;
  mmbError err = MMB_OK;
  /** Check parameters */
  if (NULL == dst_data) {
    MMB_WARN("Data parameter should not be NULL.\n");
    return MMB_INVALID_ARG;
  }
  /** forward call */
  mmb_strat_copy_1d_va_fct fwd_copy_1d = va_arg(args, mmb_strat_copy_1d_va_fct);
  void *fwd_data = va_arg(args, void *);
  err = fwd_copy_1d(dst, fwd_data, dst_interface, dst_opts, doffset,
                    src, src_interface, src_opts, soffset, width, args);
  return err;
}

mmbError mmb_strat_copy_2d_va__thread_safe(void *dst, void *dst_data,
                                           mmbMemInterface *dst_interface,
                                           const mmbAllocateOptions *dst_opts,
                                           const size_t doffset, const size_t dpitch,
                                           const void * src,
                                           const mmbMemInterface *src_interface,
                                           const mmbAllocateOptions *src_opts,
                                           const size_t soffset, const size_t spitch,
                                           const size_t width, const size_t height,
                                           va_list args)
{
  (void) dst_data;
  mmbError err = MMB_OK;
  /** Check parameters */
  if (NULL == dst_data) {
    MMB_WARN("Data parameter should not be NULL.\n");
    return MMB_INVALID_ARG;
  }
  /** forward call */
  mmb_strat_copy_2d_va_fct fwd_copy_2d = va_arg(args, mmb_strat_copy_2d_va_fct);
  void *fwd_data = va_arg(args, void *);
  err = fwd_copy_2d(dst, fwd_data, dst_interface, dst_opts, doffset, dpitch,
                    src, src_interface, src_opts, soffset, spitch,
                    width, height, args);
  return err;
}

/**  @} */

/**  @} */

/**  @} */

/**  @} */

